function [S] = engine1(fFuncs, maxTime, RT)
%simulate fight of given planes/functions
%simulation end after elapsed maxTime
%Syntax: [S] = engine1(fFuncs, maxTime, RT)
%Input parameters:
%  fFuncs is a vector of structure
%               fFuncs(k).pilot = @pilotFunc - pilot function pointer
%               fFuncs(k).color = [R G B] in range <0,1> - plane color
%  maxTime - length of fight
%  RT - simulation speed-up factor - 1 - realtime, e.g. 5 - 5x faster
%       if RT = 0, no visualization - simulation only
%Output parameters:
%  S - final score vector - a number per plane

%set time simulation difference
dt = 0.1;
if nargin < 3
    RT = 1; %real-time speed-up
end

global SIZE_X SIZE_Y
SIZE_X = 1500;
SIZE_Y = 1500;
MIN_SPEED = 100;
MAX_SPEED = 200;
MAX_BULLET_TIME = 15;
SCORE_PENALTY_PER_SHOT = 0.1;
DEATH_TIME = 30;

%init plot
if RT ~= 0
    figure(99);
    clf();
    hf = gcf(); %store handle to track figure
    set(hf, 'ToolBar', 'none');
    set(hf, 'DockControls', 'off');
    set(hf, 'Resize', 'off');
    set(hf, 'MenuBar', 'none');
    pp = get(0, 'ScreenSize'); %fot window to screensize
    pp(1) = 6;
    pp(2) = 60;
    pp(3) = pp(3) - 10;
    pp(4) = pp(4) - 80;
    set(hf, 'Position', pp);

    hx(1) = subplot(1,2,1);
    htmp1 = plot(0,0,'.');
    hx(2) = subplot(1,2,2);
    htmp2 = plot(0,0,'.');
    %hx = get(hf, 'Children');
    set(hx, 'Units', 'pixels');
    
    hold off

    set(hx, 'XTickLabel', []);
    set(hx, 'YTickLabel', []);
    set(hx, 'XTick', []);
    set(hx, 'YTick', []);
    set(hx, 'XTickMode', 'manual');
    set(hx, 'YTickLabelMode', 'manual');
    set(hx, 'XTickLabelMode', 'manual');
    set(hx, 'YTickMode', 'manual');
    set(hx, 'XLimMode', 'manual');
    set(hx, 'YLimMode', 'manual');
    set(hx, 'XLim', [0 SIZE_X]);
    set(hx, 'YLim', [0 SIZE_Y]);
    set(hx, 'Units', 'pixels');
    pp = get(hx, 'Position');
    pp{1}(3) = pp{1}(4);
    set(hx(1), 'Position', pp{1});
    pp{2}(1) = pp{1}(1)+pp{1}(3) + 100;
    pp{2}(3) = pp{1}(1)+pp{1}(3) + 400;
    set(hx(2), 'Position', pp{2}); 
    
    c = get(hf(1), 'Color');
    set(hx(2), 'Color', c);
    set(hx(2), 'XColor', c);
    set(hx(2), 'YColor', c);       
    
    delete(htmp1)
    delete(htmp2)
    
    hf = [hf hx];
    
end

Nplanes = length(fFuncs);
%prepare result matrix
S = zeros(Nplanes, 1);
bullets = [];

for k = 1:Nplanes

    %init plane
    planes(k).score = 0;
    planes(k).health = 100;
    planes(k).damage = 20;
    planes(k).xy = round([rand(1)*SIZE_X, rand(1)*SIZE_Y]);
    planes(k).angle = rand(1)*2*pi;
    planes(k).v = MIN_SPEED + rand(1)*(MAX_SPEED - MIN_SPEED);
    planes(k).throttle = 0.5;
    planes(k).steering = 0;
    planes(k).fire = 0;
    planes(k).userData = []; %user data
    planes(k).color = fFuncs(k).color;
    planes(k).death = 0;
    planes(k).plot = [];
    planes(k).setup = feval(fFuncs(k).pilot, planes(k).userData);
    if planes(k).setup == 1
        planes(k).damage = 25;
    end
    if planes(k).setup == 2
       planes(k).health = 125; 
    end

    %init plane graphics
    if RT ~= 0
        planes(k).plot = initPlanePlot(hf, planes(k), planes(k).color, k, char(fFuncs(k).pilot));
        updatePlanePlot(hf, planes(k));
    end
end

%simulation
actTime = 0;
t = 1;
while 1

    planes_old = planes;
    for k = 1:Nplanes

        if planes(k).death > 0 %skip death plane
            continue
        end
        %call the driver
        try
        [planes(k).throttle, planes(k).steering, planes(k).fire, planes(k).userData] = feval(fFuncs(k).pilot, planes_old, k, bullets, t);
        catch
            planes(k).throttle = 0;
            planes(k).steering = 0;
            planes(k).fire = 0;
        end

        %check outputs
        planes(k).throttle = max(0, min(planes(k).throttle, 1)); %0-1 limit        
        planes(k).steering = sign(planes(k).steering)*min(abs(planes(k).steering), 1); %+-1 limit        
        
        %apply top-speed boost
        planes(k).throttle = planes(k).throttle * (1 +  .75 * (planes(k).setup == 3)); % 75% throttle boost is equeal to 25% top-speed boost due to the non-linear air resistance
        %apply handling boost
        planes(k).steering = planes(k).steering * (1 +  .25 * (planes(k).setup == 4));

        
        %update firing
        if planes(k).fire ~= 0
           bullets = addBullet(bullets, planes, k, t);
           planes(k).score = planes(k).score - SCORE_PENALTY_PER_SHOT;
           if RT ~= 0 && mod(t, RT) == 0
               set(planes(k).plot(6), 'String', sprintf('% 3.1f', planes(k).score));
           end
        end

        %simulate
        planes(k) = simulateOne(planes(k), dt, MIN_SPEED);
        if RT ~= 0 && mod(t, RT) == 0
            exit = updatePlanePlot(hf, planes(k)); %move plane marker
            if exit %closed window -> exit
                disp('Window closed, simulation ends')
                return
            end
        end
    
    end %for k - Nplanes
    
    for k = 1:length(bullets)
        bullets(k) = simulateBullet(bullets(k), dt);
        if RT ~= 0 && mod(t, RT) == 0
            [bullets(k), exit] = updateBulletPlot(hf, bullets(k)); %move bullet point
            if exit %closed window -> exit
                disp('Window closed, simulation ends')
                return
            end
        end
    end
    
    %detect hits
    if ~isempty(bullets)
        [planes, bullets] = detectHits(planes, bullets, RT);
    end
    
    %respawn death planes
    f = find([planes(:).health] <= 0 | [planes(:).death] > 0);
    for k = 1:length(f)
        pid = f(k);
        if planes(pid).death == 0 %fresh death
            planes(pid).setup = feval(fFuncs(pid).pilot, planes(pid).userData);
            planes(pid).death = DEATH_TIME;
            planes(pid).health = 100;
            if planes(pid).setup == 2
                planes(pid).health = 125;
            end
            if RT ~= 0            
                set(planes(pid).plot(1), 'Visible', 'off');
                set(planes(pid).plot(2), 'Visible', 'off');
                set(planes(pid).plot(3), 'Visible', 'off');
                set(planes(pid).plot(4), 'X', [0 planes(pid).health]);
                set(planes(pid).plot(5), 'X', [planes(pid).health planes(pid).health]);
                set(planes(pid).plot(4), 'Color', [.5 .5 .5]);
            end
        else
            planes(pid).death = planes(pid).death - 1;
            if planes(pid).death == 0 %respawn the plane
                planes(pid).xy = round([rand(1)*SIZE_X, rand(1)*SIZE_Y]);
                planes(pid).angle = rand(1)*2*pi;
                planes(pid).v = MIN_SPEED + rand(1)*(MAX_SPEED - MIN_SPEED);
                if RT ~= 0                
                    set(planes(pid).plot(1), 'Visible', 'on');
                    set(planes(pid).plot(2), 'Visible', 'on');
                    set(planes(pid).plot(3), 'Visible', 'on');
                    set(planes(pid).plot(4), 'Color', [0 .5 0]);
                    updatePlanePlot(hf, planes(pid));
                end
            end
        end
    end

    %test end of fight
    if actTime >= maxTime
        break %end of fight
    end
    
    if RT ~= 0
        pause(dt/RT);
    end
    t = t + 1;
    actTime = actTime + dt;
    
    %discard old bullets
    if ~isempty(bullets)
        ob = t - [bullets(:).time] > MAX_BULLET_TIME;
        if RT ~= 0 %delete bullet plots
            try
                delete([bullets(ob).handle]);
            catch
                disp('Window closed, simulation ends')
                break
            end
        end
        bullets(ob) = [];
    end
   
end %while 1
if RT ~= 0
	disp('Simulation end')
end
S = [planes(:).score];

end %main function

function plotHandles = initPlanePlot(hf, plane, color, id, driver)

set(hf(1),'DoubleBuffer','on')
axes(hf(2));
hold on
plotHandles(1) = plot(plane.xy(1), plane.xy(2), 'LineWidth', 2, 'Color', color);
plotHandles(2) = plot(plane.xy(1), plane.xy(2), 'LineWidth', 4, 'Color', color);
plotHandles(3) = plot(plane.xy(1), plane.xy(2), 'LineWidth', 4, 'Color', color);
hold off

%legend:
axes(hf(3));
hold on
hx = hf(3);
step = 150;
%px = get(hx, 'Position');
X = 0;
px = get(hx, 'YLim');
Y = px(2) + step - 20;

plotHandles(4) = plot([X X + plane.health], [Y - 30 - id*step, Y - 30 - id*step], 'LineWidth', 5, 'Color', [0 .5 0]); %health bar green part
plotHandles(5) = plot([X + plane.health, X + plane.health], [Y - 30 - id*step, Y - 30 - id*step], 'LineWidth', 5, 'Color', [.5 0 0]); %health bar red part
text(X, Y - id * step, driver);
plot([X X + 60], [Y - 70 - id*step, Y - 70 - id*step], 'LineWidth', 4, 'Color', color); %legend-plane
plot([X + 30 X + 30], [Y - 40 - id*step, Y - 100 - id*step], 'LineWidth', 2, 'Color', color); %legend-plane
plot([X + 15 X + 45], [Y - 95 - id*step, Y - 95 - id*step], 'LineWidth', 4, 'Color', color); %legend-plane
plotHandles(6) = text(X + 70, Y - 80 - id * step, sprintf('% 3.1f', plane.score));
set(plotHandles(6), 'FontSize', 18);

hold off

end %initPlanePlot

function exit = updatePlanePlot(hf, plane)

exit = 0;
try
    hh = gcf(); %test window close
    if hh ~= hf
       close(hh);
       exit = 1;
       return
    end
figure(hf(1)); %set hf-handle figure as current

X = [plane.xy(1) - cos(plane.angle) * 30, plane.xy(1) + cos(plane.angle) * 30]; %absolute positions X
Y = [plane.xy(2) - sin(plane.angle) * 30, plane.xy(2) + sin(plane.angle) * 30]; %absolute positions Y
set(plane.plot(1), 'X', X);
set(plane.plot(1), 'Y', Y);

X = [plane.xy(1) - cos(pi/2+plane.angle) * 30, plane.xy(1) + cos(pi/2+plane.angle) * 30]; %absolute positions X
Y = [plane.xy(2) - sin(pi/2+plane.angle) * 30, plane.xy(2) + sin(pi/2+plane.angle) * 30]; %absolute positions Y
set(plane.plot(2), 'X', X);
set(plane.plot(2), 'Y', Y);

X = [plane.xy(1) + cos(plane.angle+2.7) * 35, plane.xy(1) + cos(plane.angle-2.7) * 35]; %absolute positions X
Y = [plane.xy(2) + sin(plane.angle+2.7) * 35, plane.xy(2) + sin(plane.angle-2.7) * 35]; %absolute positions Y
set(plane.plot(3), 'X', X);
set(plane.plot(3), 'Y', Y);

catch
    exit = 1;
    return
end
end %updatePlanePlot


function plane = simulateOne(plane, dt, MIN_SPEED)

plane.angle = unwrapAngle(plane.angle + dt*plane.steering); %update angle according to steering and speed
plane.v = plane.v*cos(plane.steering)^dt; %slowdown caused by steering
resistance =  -100 + .01*plane.v^2;
acel = dt*(300*plane.throttle - resistance);
plane.v = max(plane.v + acel, MIN_SPEED); %update speed
vxy = plane.v * [cos(plane.angle), sin(plane.angle)];
plane.xy = plane.xy + dt*vxy; %update position

%unwrap xy
plane.xy = unwrapXY(plane.xy);

end %simulateOne


function bullets = addBullet(bullets, planes, k, t)
    bullets(end+1).id = k; %who was firing
    bullets(end).time = t; %when
    bullets(end).xy = planes(k).xy + 30*[cos(planes(k).angle), sin(planes(k).angle)]; %where
    bullets(end).angle = planes(k).angle; %which way
    bullets(end).v = planes(k).v + 200; %what speed
    bullets(end).handle = []; %handle to graphical point
    bullets(end).color = planes(k).color;
    bullets(end).hitpoints = planes(k).damage;
end %addBullet

function bullet = simulateBullet(bullet, dt)

    vxy = bullet.v * [cos(bullet.angle), sin(bullet.angle)];
    bullet.xy = bullet.xy + dt*vxy; %update position

    %unwrap xy
    bullet.xy = unwrapXY(bullet.xy);

end %simulateBullet


function [bullet, exit] = updateBulletPlot(hf, bullet)
    
    exit = 0;
    try
        hh = gcf(); %test window close
        if hh ~= hf
           close(hh);
           exit = 1;
           return
        end
    figure(hf(1)); %set hf-handle figure as current

    %if first plot
    if isempty(bullet.handle)
        axes(hf(2));
        hold on
        bullet.handle = plot(bullet.xy(1), bullet.xy(2), '.', 'LineWidth', 3, 'Color', bullet.color);
        hold off
    else
        set(bullet.handle, 'X', bullet.xy(1));
        set(bullet.handle, 'Y', bullet.xy(2));
    end
        
    catch
        exit = 1;
        return
    end
end %updateBulletPlot


function [planes, bullets] = detectHits(planes, bullets, RT)
    P = length(planes);
    B = length(bullets);

    %get planes x and y as vectors
    pxy = [planes(:).xy];
    px = pxy(1:2:end)';
    py = pxy(2:2:end)';
    %get byllets x and y as vectors    
    bxy = [bullets(:).xy];
    bx = bxy(1:2:end);
    by = bxy(2:2:end);

    %compute distance matrix
    D = sqrt((repmat(px, 1, B) - repmat(bx, P, 1)).^2 + (repmat(py, 1, B) - repmat(by, P, 1)).^2);
    D([planes(:).death] > 0, :) = Inf; %dont hit the death ones
    %find minimum distance from bullet
    [mD, mDi] = min(D);
    
    f = find(mD < 30); %detect the hit

    if ~isempty(f)
       for k = 1:length(f) 
           bid = f(k);
           pid = mDi(bid);
           poid = bullets(bid).id;
           planes(pid).health = max(planes(pid).health - bullets(bid).hitpoints, 0);
           planes(poid).score = planes(poid).score + 1 + (planes(pid).health <= 0) + 0.25 * (planes(poid).setup == 1); % +25% if stronger weapon, +100% if kill
           if RT ~= 0
              set(planes(pid).plot(4), 'X', [0  planes(pid).health]);
              set(planes(pid).plot(5), 'X', [planes(pid).health 100 + 25*(planes(pid).setup==2)]);
              set(planes(poid).plot(6), 'String', sprintf('% 3.1f', planes(poid).score));
           end
       end
    end
    bh = [bullets(f).handle];
    %delete(bh(~isempty(bh)));
    if ~isempty(bh)
        delete(bh);
    end
    bullets(f) = [];
end %detectHits
